/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_kgdb.c,v $
 *
 * HISTORY
 * $Log: alpha_kgdb.c,v $
 * Revision 1.8  1994/06/11  00:02:38  menze
 * Added copyright notice
 *
 * Revision 1.7  1994/05/14  23:39:13  davidm
 * (kgdb_handle_exception): disable interrupts while in GDB---makes it easier
 * to debug things.
 *
 * Revision 1.6  1994/04/30  22:56:22  davidm
 * P command fixed to use `=' instead of `:'
 *
 * Don't use bcopy() to reduce probability of endless recursion.
 *
 * Revision 1.5  1994/03/29  00:22:05  davidm
 * Various minor fixes.
 *
 * Revision 1.4  1994/03/27  17:03:10  davidm
 * (mem2hex,hex2mem): Fixed such that they work---it is not wise to
 * access memory that may fault by byte---one access may translate into
 * several memory accesses, defeating the "scout_expect_memory_access_fault"
 * mechanism.
 *
 * (regno2addr): New function.
 *
 * (handle_exception): Support for writing individual registers added (P
 * command).  Also, FP is now included in the T response---GDB wants it.
 *
 * (kgdb_handle_exception): Function kgdb_restore_and_rti() is now used to
 * return from an exception reliably.
 *
 * Revision 1.3  1994/03/26  01:55:33  davidm
 * (kgdb_set_debug_traps): no longer invokes kgdb_breakpoint().
 *
 * Revision 1.2  1994/03/17  00:14:42  davidm
 * (handle_exception): added a comment.
 *
 * Revision 1.1  1994/03/16  22:59:06  davidm
 * Initial revision
 *
 */
/*
 * This module provides low-level support for GDB and has been adapted
 * from sparc-stub.c.  Please refer to that file for more on the
 * history of the module.  The contributors seem to be: Hewlett-Packard,
 * Glenn Engel, and Cygnus Support (Stu Grossman).
 *
 * To enable debugger support, two things need to happen.  One, a call
 * to kgdb_set_debug_traps() is necessary in order to allow any
 * breakpoints or error conditions to be properly intercepted and
 * reported to gdb.  Two, a breakpoint needs to be generated to begin
 * communication.  This is most easily accomplished by a call to
 * kgdb_breakpoint().  kgdb_breakpoint() simulates a breakpoint by
 * executing a call_pal bpt.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *    Pn..=v..	    set CPU register n.. to value v..	   OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: <two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 */

#include <string.h>
#include <signal.h>
#include <memory.h>
#include <machine/pal.h>
#include "scout.h"
#include "scout_sync_arch.h"
#include "alpha_misc.h"
#include "kgdb.h"

#define initDebugIO	scc_attach
#define putDebugChar	scc_putc
#define getDebugChar	scc_getc

/*
 * external low-level support routines
 */
extern void initDebugIO (void);		/* initialize debugging I/O	 */
extern void putDebugChar (int ch);	/* write a single character      */
extern int  getDebugChar (void);	/* read and return a single char */

#define BREAKINST	0x00000080	/* call_pal bpt */

/*
 * BUFMAX defines the maximum number of characters in inbound/outbound
 * buffers at least NUMREGBYTES*2 are needed for register packets.
 */
#define BUFMAX 2048

#define NUMREGS		(2*32+1)	/* # of registers */
#define NUMREGBYTES	(NUMREGS * 8)	/* # of bytes for registers */

enum regnames {
      V0,   T0,   T1,   T2,   T3,   T4,   T5,   T6,
      T7,   S0,   S1,   S2,   S3,   S4,   S5,   S6,
      A0,   A1,   A2,   A3,   A4,   A5,   T8,   T9,
     T10,  T11,   RA,  T12,   AT,   GP,   SP, ZERO,
      F0,   F1,   F2,   F3,   F4,   F5,   F6,   F7,
      F8,   F9,  F10,  F11,  F12,  F13,  F14,  F15,
     F16,  F17,  F18,  F19,  F20,  F21,  F22,  F23,
     F24,  F25,  F26,  F27,  F28,  F29,  F30,  F31,
      PC,  VFP
};

#define DBG_STACK		0x01
#define DBG_TRAP		0x02
#define DBG_IN_PACKETS		0x04
#define DBG_OUT_PACKETS		0x08
#define DBG_MEM			0x10
#define DBG_STORE_REG		0x20
#define DBG_CMDS		0x40
#define DBG_SINGLE_STEP		0x80

#ifdef SCOUT_DEBUG
# define DBG(bit,args)	{if (kgdb_debug & (bit)) printf args;}
#else
# define DBG(bit,args)
#endif

int kgdb_debug = 0;		/* kgdb debugging level */

static const char hexchars[] = "0123456789abcdef";

static int	nsaved = 0;
static u_int	saved_insn[2];
static u_int	*saved_addr[2];

static int initialized = 0;	/* !0 means we've been initialized */
static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/*
 * This table contains the mapping between Alpha hardware trap types,
 * and signals, which are primarily what GDB understands.  It also
 * indicates which hardware traps we need to commandeer when
 * initializing the stub.
 */
static const struct hard_trap_info
{
    u_char tt;		/* Trap type code for Alpha */
    u_char signo;	/* Signal that we map this trap into */
} hard_trap_info[] = {
    {0, SIGTRAP},	/* breakpoint instruction */
    {1, SIGILL},	/* illegal instruction (bug check) */
    {2, SIGEMT},	/* gentrap */
    {3, SIGFPE},	/* fp disabled */
    {4, SIGSYS},	/* opDEC fault */
    {0xff, 0xff}	/* must be last */
};


/*
 * Don't want to use libraries bcopy() to avoid infinite recursion:
 */
static void
private_bcopy(void *from, void *to, int length)
{
    char *fp = from;
    char *tp = to;

    while (length--) *tp++ = *fp++;
} /* private_bcopy */


/*
 * Convert ch from a hex digit to an int:
 */
static int
hex(u_char ch)
{
    if (ch >= 'a' && ch <= 'f') {
	return ch - 'a' + 10;
    } else if (ch >= '0' && ch <= '9') {
	return ch - '0';
    } else if (ch >= 'A' && ch <= 'F') {
	return ch - 'A' + 10;
    } else {
	return -1;
    } /* if */
} /* hex */


/*
 * Scan for the sequence $<data>#<checksum>:
 */
static void
getpacket(char *buffer)
{
    u_char checksum;
    u_char xmitcsum;
    int i;
    int count;
    u_char ch;

    do {
	/* wait around for the start character, ignore all other characters */
	while ((ch = getDebugChar()) != '$') {
	    /* skip */
	} /* while */

	checksum = 0;
	xmitcsum = -1;

	count = 0;

	/* now, read until a # or end of buffer is found */
	while (count < BUFMAX) {
	    ch = getDebugChar();
	    if (ch == '#') {
		break;
	    } /* if */
	    checksum = checksum + ch;
	    buffer[count] = ch;
	    count = count + 1;
	} /* while */

	if (count >= BUFMAX) {
	    continue;
	} /* if */

	buffer[count] = 0;

	if (ch == '#') {
	    xmitcsum = hex(getDebugChar()) << 4;
	    xmitcsum |= hex(getDebugChar());
	    if (checksum != xmitcsum) {
		putDebugChar('-');	/* failed checksum */
	    } else {
		putDebugChar('+'); /* successful transfer */
		/* if a sequence char is present, reply the sequence ID */
		if (buffer[2] == ':') {
		    putDebugChar(buffer[0]);
		    putDebugChar(buffer[1]);
		    /* remove sequence chars from buffer */
		    count = strlen(buffer);
		    for (i=3; i <= count; i++) {
			buffer[i-3] = buffer[i];
		    } /* for */
		} /* if */
	    } /* if */
	} /* if */
    } while (checksum != xmitcsum);

    DBG(DBG_IN_PACKETS, ("kgdb: got packet `$%s'\n", remcomInBuffer));
} /* getpacket */


/*
 * Send the packet in buffer.
 */
static void
putpacket(u_char *buffer)
{
    u_char checksum;
    int count;
    u_char ch;

    DBG(DBG_OUT_PACKETS, ("kgdb: sending packet `%s'\n", remcomOutBuffer));

    /*  $<packet info>#<checksum>. */
    do {
	putDebugChar('$');
	checksum = 0;
	count = 0;

	while ((ch = buffer[count]) != '\0') {
	    putDebugChar(ch);
	    checksum += ch;
	    count += 1;
	} /* while */

	putDebugChar('#');
	putDebugChar(hexchars[checksum >> 4]);
	putDebugChar(hexchars[checksum & 0xf]);

    } while (getDebugChar() == '-');
} /* putpacket */


/*
 * Convert the memory pointed to by mem into hex, placing result in
 * buf.  Return a pointer to the last char put in buf (null), in case
 * of mem fault, return 0.  If MAY_FAULT is non-zero, then we will
 * handle memory faults by returning a 0, else treat a fault like any
 * other fault in the stub.
 */
static u_char *
mem2hex(u_char *mem, u_char *buf, int count, int may_fault)
{
    u_char ch;
    u_long l, *lp;
    int byte;
    
    lp = (u_long*) ((u_long) mem & ~0x7);
    byte = mem - (u_char*) lp;
    scout_expect_memory_access_fault = TRUE;
    l = *lp++;
    l >>= 8 * byte;
    while (count-- > 0) {
	if (!scout_expect_memory_access_fault) {
	    return 0;
	} /* if */
	ch = l & 0xff;
	*buf++ = hexchars[ch >> 4];
	*buf++ = hexchars[ch & 0xf];
	l >>= 8;
	if (++byte >= 8) {
	    byte = 0;
	    l = *lp++;
	} /* if */
    } /* while */
    scout_expect_memory_access_fault = FALSE;
    *buf = 0;
    return buf;
} /* mem2hex */


/*
 * Convert the hex array pointed to by buf into binary to be placed in
 * mem return a pointer to the character AFTER the last byte written
 */
static char *
hex2mem(u_char *buf, u_char *mem, int count, int may_fault)
{
    u_char ch;
    u_long l, *lp;
    int byte;
    int i;
    
    lp = (u_long*) ((u_long) mem & ~0x7);
    byte = mem - (u_char*) lp;
    scout_expect_memory_access_fault = TRUE;
    l = *lp;
    for (i = 0; i < count; ++i) {
	if (!scout_expect_memory_access_fault) {
	    return 0;
	} /* if */
	ch = hex(*buf++) << 4;
	ch |= hex(*buf++);
	l = (l & ~(0xffL << 8*byte)) | ((long) ch << 8*byte);
	if (++byte >= 8) {
	    byte = 0;
	    *lp++ = l;
	    if (!scout_expect_memory_access_fault) {
		return 0;
	    } /* if */
	    l = *lp;
	} /* if */
    } /* for */
    if (byte) {
	*lp = l;
    } /* if */
    scout_expect_memory_access_fault = FALSE;
    return buf;
} /* hex2mem */


/*
 * Convert the Alpha hardware trap type code to a unix signal number.
 */
static int
compute_signal(int tt)
{
    const struct hard_trap_info *ht;

    for (ht = hard_trap_info; ht->tt < 0xff ; ht++) {
	if (ht->tt == tt) {
	    return ht->signo;
	} /* if */
    } /* for */

    return SIGHUP;		/* default for things we don't know about */
} /* compute_signal */


/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */
static int
hex2int(char **ptr, int *int_value)
{
    int num_chars = 0;
    int hex_value;

    *int_value = 0;

    while (**ptr) {
	hex_value = hex(**ptr);
	if (hex_value < 0) {
	    break;
	} /* if */

	*int_value = (*int_value << 4) | hex_value;
	num_chars ++;

	(*ptr)++;
    } /* while */

    return num_chars;
} /* hex2int */


/*
 * While we find nice hex chars, build a long.
 * Return number of chars processed.
 */
static int
hex2long(char **ptr, long *long_value)
{
    int num_chars = 0;
    long hex_value;

    *long_value = 0;

    while (**ptr) {
	hex_value = hex(**ptr);
	if (hex_value < 0) {
	    break;
	} /* if */

	*long_value = (*long_value << 4) | hex_value;
	num_chars ++;

	(*ptr)++;
    } /* while */

    return num_chars;
} /* hex2long */


/*
 * Convert a register number to the corresponding pointer into
 * the saved state.  Returns 0 in the case of a bad register.
 */
static u_long*
regno2addr(u_int regno, Alpha_Saved_State *as, u_long **sp)
{
    static u_long zero;
    u_long *addr;

    switch (regno) {
      case V0...S6: addr = &as->v0 + regno; break;
      case A3...AT: addr = &as->a3 + regno - A3; break;
      case A0...A2: addr = &as->a0 + regno - A0; break;
      case SP:	    addr = (u_long*) sp; break;
      case PC:	    addr = (u_long*) &as->pc; break;
      case GP:	    addr = &as->gp; break;
      case ZERO:    addr = &zero; zero = 0L; break;
      default:	    addr = 0; break;
    } /* switch */
    return addr;
} /* regno2addr */


static int
set_breakpoint(Alpha_Saved_State *as, u_long *sp)
{
    int op_code, displ, reg_b, i;
    u_long *reg_b_addr;
    u_int insn;

    scout_expect_memory_access_fault = TRUE;
    insn = *as->pc;
    if (!scout_expect_memory_access_fault) {
	return FALSE;
    } /* if */
    scout_expect_memory_access_fault = FALSE;

    op_code = insn >> 26;
    DBG(DBG_SINGLE_STEP, ("kgdb: single-step over "));
    if (op_code >= 0x30) {
	/*
	 * It's a branch: instead of trying to figure out whether the
	 * branch will be taken or not, we'll put a breakpoint at either
	 * location.  This is simpler, more reliable, and probably not
	 * a whole lot slower than the alternative approach of emulating
	 * the branch (emulation can be tricky for fp branches).
	 */
	DBG(DBG_SINGLE_STEP, ("branch"));
	displ = ((int)(insn<<11)) >> 11;
	saved_addr[nsaved++] = as->pc + 1;
	saved_addr[nsaved++] = as->pc + 1 + displ;
    } else if (op_code == 0x1a) {
	/* it's a jump */
	DBG(DBG_SINGLE_STEP, ("jump"));
	reg_b = (insn >> 16) & 0x1f;
	reg_b_addr = regno2addr(reg_b, as, &sp);
	if (reg_b_addr) {
	    saved_addr[nsaved++] = (u_int*) *reg_b_addr;
	} else {
	    return FALSE;
	} /* if */
    } else {
	DBG(DBG_SINGLE_STEP, ("normal"));
	saved_addr[nsaved++] = as->pc + 1;
    } /* if */

    for (i = 0; i < nsaved; ++i) {
	scout_expect_memory_access_fault = TRUE;
	saved_insn[i] = *saved_addr[i];
	if (!scout_expect_memory_access_fault) {
	    nsaved = 0;
	    return FALSE;
	} /* if */
	DBG(DBG_SINGLE_STEP, (" -> next_pc=%p", saved_addr[i]));
	*saved_addr[i] = BREAKINST;
	if (!scout_expect_memory_access_fault || *saved_addr[i] != BREAKINST) {
	    nsaved = 0;
	    return FALSE;
	} /* if */
	scout_expect_memory_access_fault = FALSE;
    } /* for */
    DBG(DBG_SINGLE_STEP, ("\n"));
    return TRUE;
} /* set_breakpoint */


/*
 * This function does all command processing for interfacing to gdb.  It
 * returns the new stack pointer.
 */
static u_long*
handle_exception(long type, Alpha_Saved_State *as, u_long *sp)
{
    int sigval;
    u_int regno;
    int length;
    char *ptr;
    u_long addr, *reg_addr;

    if (type == 0 && as->pc[-1] != BREAKINST) {
	--as->pc;		/* make pc point to restored insn */
    } /* if */

    DBG(DBG_TRAP, ("kgdb: handle_exception(type=%ld, as=%p, sp=%p, pc=%p)\n",
		   type, as, sp, as->pc));

    /*
     * Reply to host that an exception has occurred.  Include the
     * minimal state GDB needs in order to do anything useful.  The
     * easiest way to find out which registers should be sent is to
     * add a printf to function read_register() in gdb/findvar.c and
     * do an "si" after connecting to the stub.
     */
    sigval = compute_signal(type);
    ptr = remcomOutBuffer;
    *ptr++ = 'T';
    *ptr++ = hexchars[sigval >>  4];
    *ptr++ = hexchars[sigval & 0xf];

    *ptr++ = hexchars[S6 >>  4];	/* s6 also serves as frame pointer */
    *ptr++ = hexchars[S6 & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&as->s6, ptr, 8, 0);
    *ptr++ = ';';

    *ptr++ = hexchars[SP >>  4];
    *ptr++ = hexchars[SP & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&sp, ptr, 8, 0);
    *ptr++ = ';';

    *ptr++ = hexchars[PC >>  4];
    *ptr++ = hexchars[PC & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&as->pc, ptr, 8, 0);
    *ptr++ = ';';

    /*
     * GDB wants VFP --- even though it's a computed value:
     */
    *ptr++ = hexchars[VFP >>  4];
    *ptr++ = hexchars[VFP & 0xf];
    *ptr++ = ':';
    memset(ptr, '0', 8 * 2); ptr += 8 * 2;
    *ptr++ = ';';

    *ptr++ = '\0';
    DBG(DBG_OUT_PACKETS, ("kgdb: sending trap info `%s'\n", remcomOutBuffer));
    putpacket(remcomOutBuffer);

    while (1) {
	remcomOutBuffer[0] = '\0';

	getpacket(remcomInBuffer);
	DBG(DBG_CMDS, ("kgdb: got command `%c'\n", remcomInBuffer[0]));
	switch (remcomInBuffer[0]) {
	  case '?':	/* return last signal */
	    remcomOutBuffer[0] = 'S';
	    remcomOutBuffer[1] = hexchars[sigval >> 4];
	    remcomOutBuffer[2] = hexchars[sigval & 0xf];
	    remcomOutBuffer[3] = 0;
	    break;

	  case 'd':	/* toggle debug flag */
	    break;

	  case 'g':	/* return the value of the CPU registers */
	    ptr = remcomOutBuffer;
	    ptr = mem2hex((char*) &as->v0, ptr, 16 * 8, 0); /* v0..s6 */
	    ptr = mem2hex((char*) &as->a0, ptr,  3 * 8, 0); /* a0..a2 */
	    ptr = mem2hex((char*) &as->a3, ptr, 10 * 8, 0); /* a3..at */
	    ptr = mem2hex((char*) &as->gp, ptr,  1 * 8, 0); /* gp */
	    ptr = mem2hex((char*) &sp, ptr,  1 * 8, 0); /* sp */
	    memset(ptr, '0', 33 * 8 * 2);		    /* zero, f0..f31 */
	    ptr += 33 * 8 * 2;
	    ptr = mem2hex((char*) &as->pc, ptr,  1 * 8, 0); /* pc */
	    break;

	  case 'G':	/* set the value of the CPU registers - return OK */
	    ptr = &remcomInBuffer[1];
	    ptr = hex2mem(ptr, (char*) &as->v0, 16 * 8, 0); /* v0..fp */
	    ptr = hex2mem(ptr, (char*) &as->a0,  3 * 8, 0); /* a0..a2 */
	    ptr = hex2mem(ptr, (char*) &as->a3, 10 * 8, 0); /* a3..at */
	    ptr = hex2mem(ptr, (char*) &as->gp,  1 * 8, 0); /* gp */
	    ptr = hex2mem(ptr, (char*) &sp,  1 * 8, 0); /* sp */
	    ptr += 33 * 8 * 2;				    /* zero, f0..f31 */
	    ptr = hex2mem(ptr, (char*) &as->pc,  1 * 8, 0); /* pc */
	    strcpy(remcomOutBuffer, "OK");
	    break;

	  case 'P':	/* set the value of one CPU register - return OK */
	    ptr = &remcomInBuffer[1];
	    if (hex2int(&ptr, &regno) && *ptr++ == '=' && regno <= VFP) {
		reg_addr = regno2addr(regno, as, &sp);
		DBG(DBG_STORE_REG, ("kgdb: storing %s to reg %d @ %p "
				    "(as=%p, &sp=%p)\n", ptr, regno, reg_addr,
				    as, &sp));
		if (reg_addr) {
		    hex2mem(ptr, (char*)reg_addr, 8, 0);
		} /* if */
		strcpy(remcomOutBuffer, "OK");
	    } else {
		strcpy(remcomOutBuffer, "E01");
	    } /* if */
	    break;

	  case 'm':	/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr) && *ptr++ == ','
		&& hex2int(&ptr, &length))
	    {
		DBG(DBG_MEM, ("kgdb: reading %d bytes at %lx\n",
			      length, addr));
		if (mem2hex((char *)addr, remcomOutBuffer, length, 1)) {
		    break;
		} /* if */
		strcpy(remcomOutBuffer, "E03");
	    } else {
		strcpy(remcomOutBuffer, "E01");
	    } /* if */
	    break;

	  case 'M':
	    /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */

	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr)
		&& *ptr++ == ','
		&& hex2int(&ptr, &length)
		&& *ptr++ == ':')
	    {
		DBG(DBG_MEM, ("kgdb: writing %d bytes %s at %lx\n", length,
			      ptr, addr));
		if (hex2mem(ptr, (char *)addr, length, 1)) {
		    strcpy(remcomOutBuffer, "OK");
		} else {
		    strcpy(remcomOutBuffer, "E03");
		} /* if */
	    } else {
		strcpy(remcomOutBuffer, "E02");
	    } /* if */
	    break;

	  case 's':	/* sAA..AA    Step at address AA..AA (optional) */
	  case 'c':	/* cAA..AA    Continue at address AA..AA (optional) */
	    /* try to read optional parameter, pc unchanged if no parm: */
	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr)) {
		as->pc = (u_int*) addr;
	    } /* if */
	    DBG(DBG_SINGLE_STEP, ("kgdb: continue/step (%c) at %p\n",
				  remcomInBuffer[0], as->pc));
	    if (remcomInBuffer[0] == 's' && !set_breakpoint(as, sp)) {
		strcpy(remcomOutBuffer, "E03");
		break;
	    } /* if */
	    /*
	     * Need to flush the instruction cache here, as we may
	     * have deposited a breakpoint, and the icache has no way
	     * of knowing that a data ref to some location may have
	     * changed something that is in the instruction cache.
	     */
	    flush_icache();
	    return sp;


	  case 'k':	/* kill the program */
	    break;

	  case 'r':	/* reset */
	    halt();
	    break;

	  default:
	    printf("kgdb: handle_exception: unknown command `%c'\n",
		   remcomInBuffer[0]);
	    break;
	} /* switch */

	/* reply to the request */
	putpacket(remcomOutBuffer);
    } /* while */
} /* handle_exception */


/*
 * Set up exception handlers for tracing and breakpoints.
 */
void
kgdb_set_debug_traps()
{
    initialized = 1;
    initDebugIO();
} /* kgdb_set_debug_traps */


/*
 * The following is the nasty part of entering KGDB.  GDB assumes that
 * it can access the memory below the interrupted thread's stack
 * pointer (SP).  This means that the stub cannot run on this stack
 * (interrupts could overwrite changes made by GDB).  Similarly, we do
 * not want to bother the GDB user with the call chain leading from
 * tne interrupt handler to this function.  That is, the SP reported
 * to GDB is the thread's SP at the point it was interrupted.  We
 * achieve all this by copying the saved state and the stack frames
 * below it to a private KGDB stack.  We then switch to that stack;
 * GDB does whatever it wants to the thread's stack; and when it
 * continues execution, we restore the registers from the copy of the
 * saved state and return from the interrupt.
 */
void
kgdb_handle_exception(long type, void *saved_state)
{
    static u_long kgdb_stack[1024];
    static u_long *sp = &kgdb_stack[1024];
    register u_long *reg_sp asm ("30");
    u_long *saved_sp, *new_sp;
    Alpha_Saved_State *as;
    int nlongs, i;
    extern void kgdb_restore_and_rti(Alpha_Saved_State *as, u_long *new_sp);

    asm volatile ("" ::: "30");	/* make GCC believe that sp is defined... */

    /*
     * Disable interrupts while in debugger (old IPL will be restored
     * via RTI):
     */
    scout_swap_ipl(SCOUT_MAX_IPL);

    /*
     * Restore original instructions as quickly as possible---in the
     * case the stub executes the same code.
     */
    for (i = 0; i < nsaved; ++i) {
	/* restore original instruction */
	DBG(DBG_SINGLE_STEP, ("kgdb: single-step restoring instruction %x "
			      "at %p\n", saved_insn[i], saved_addr[i]));
	*saved_addr[i] = saved_insn[i];
    } /* for */
    nsaved = 0;

    saved_sp = (u_long*)(saved_state + sizeof(Alpha_Saved_State));

    /* copy saved state to exception stack: */
    nlongs = saved_sp - reg_sp;
    as = (Alpha_Saved_State*) (sp - sizeof(Alpha_Saved_State)/sizeof(u_long));
    sp -= nlongs;
    if (sp <= kgdb_stack) {
	new_sp = sp;
	sp = kgdb_stack;
	printf("kgdb: stack underflow (sp=%p, kgdb_stack=%p)!\n",
	       new_sp, kgdb_stack);
    } /* if */
    private_bcopy(reg_sp, sp, nlongs * sizeof(u_long));

    DBG(DBG_STACK, ("kgdb_handle_exception: stack at %p\n", sp));

    /* switch to exception stack: */
    asm volatile ("mov %0,$30" :: "r"(sp) : "memory");

    new_sp = handle_exception(type, as, saved_sp);

    /* switch back to process's stack---which might have moved: */
    DBG(DBG_STACK, ("switching stack to %p (saved_sp=%p)\n",
		    new_sp, saved_sp));

    sp += nlongs;	/* pop stack */

    kgdb_restore_and_rti(as, new_sp);
} /* kgdb_handle_exception */


/*
 * This function will generate a breakpoint exception.  It is used at
 * the beginning of a program to sync up with a debugger and can be
 * used otherwise as a quick means to stop program execution and
 * "break" into the debugger.
 */
void
kgdb_breakpoint(void)
{
    if (!initialized) {
	return;
    } /* if */
    asm("call_pal %0" :: "i"(PAL_bpt));
} /* kgdb_breakpoint */

			/*** end of kgdb-alpha.c ***/
